-- Codename Dream9 by Basil Voronkov
-- Module 'drawing'
---------------------------------------------
-- This module contains various drawing and
-- measurement functions. It is based on 'sprites'
-- module.
require "sprites"
require "lib/const"
require "lib/core"
require "lib/string"

drawing = {
  fontcache = {},
  spritecache = {}
}

function drawing.init()
  --clear all cache
  for k,v in ipairs(drawing.fontcache) do
    sprite.free_font(v);
    drawing.fontcache[k] = nil;
  end
  for k,v in ipairs(drawing.spritecache) do
    sprite.free(v);
    drawing.spritecache[k] = nil;
  end
  drawing.fontcache={};
  drawing.spritecache={};
end

function drawing.font(name, size)
  if name == nil then
    name = const.font;
  end
  if size == nil then
    size = const.fontsize;
  end
  local key = name..tostring(size);
  local f = drawing.fontcache[key];
  if f == nil then
    f = sprite.font(const.path.theme..name, size);
    drawing.fontcache[key] = f;
  end
  return f;
end

function drawing.text_height(fn, fs)
  local f = drawing.font(fn, fs);
  return math.ceil(sprite.font_scaled_size(sprite.font_height(f)));
end

function drawing.text_width(txt, fn, fs)
  local f = drawing.font(fn, fs);
  return math.ceil(sprite.font_scaled_size(sprite.text_size(f, txt)));
end

function drawing.text_bounds(txt, width, fn, fs)
  local function calc_rows(txt)
    if txt == nil then
      return 0;
    end  
    local ln = txt:lines();
    local rows = 0;
    for i,e in pairs(ln) do
      if e == "" then
        rows = rows + 1;
      else
        local rl = drawing.text_width(e, fn, fs);
        local n = math.ceil(rl / width);
        rows = rows + n;
      end
    end    
    return rows;
  end
  local r = calc_rows(txt,fn);
  local fh = drawing.text_height(fn, fs);
  return fh*r;
end

function drawing.sprite_blank(w, h, cache)
  if cache then
    local key = w.."+"..h;
    local spr = drawing.spritecache[key];
    if spr == nil then
      spr = sprite.blank(w, h);
      drawing.spritecache[key] = spr;
    end
    return spr;
  else
    return sprite.blank(w, h);
  end
end

function drawing.box(w, h, c, o)
  local key = w.."+"..h.."+"..c.."+"..o;
  local spr = drawing.spritecache[key];
  if spr == nil then
    spr = sprite.box(w, h, c, o);
    drawing.spritecache[key] = spr;
  end
  return spr;
end

function drawing.sprite(file)
  local spr = drawing.spritecache[file];
  if spr == nil then
    spr = sprite.load(file);
    drawing.spritecache[file] = spr;
  end
  return spr;
end

function drawing.clean(pic, x, y, w, h)
  notnil(w, "w");
  notnil(h, "h");
  local pt = drawing.sprite_blank(w, h);
  sprite.copy(pt, pic, x, y);
  sprite.free(pt);
end

function drawing.copy(src, dst, x, y, alpha)
  local w,h = sprite.size(src);
  sprite.compose(src, 0, 0, w, h, dst, x, y, alpha);
end

function drawing.draw_mask_generic(pic, x, y, w, h, c, o)
  if c == nil then
    c = const.maskcolor;
  end
  if o == nil then
    o = const.opacity;
  end
  local b = drawing.box(w, h, c, o);
  sprite.draw(b, pic, x, y);
end

function drawing.draw_mask(pic, x, y, w, h, c, o)
  drawing.draw_mask_generic(pic, x, y, w, h, c, o);
end

function drawing.draw_sel(pic, x, y, w, h, c, o)
  drawing.draw_mask_generic(pic, x, y, w, h, c, o);
end

function drawing.copy_mask_generic(pic, x, y, w, h, c, o)
  if c == nil then
    c = const.maskcolor;
  end
  if o == nil then
    o = const.opacity;
  end
  local b = drawing.box(w, h, c, o);
  drawing.copy(b, pic, x, y);
end

function drawing.copy_mask(pic, x, y, w, h, c, o)
  drawing.copy_mask_generic(pic, x, y, w, h, c, o);
end

function drawing.copy_sel(pic, x, y, w, h, o)
  drawing.copy_mask_generic(pic, x, y, w, h, const.selcolor, o);
end

function drawing.render_text(copy, pic, txt, x, y, c, fn, fs, alpha)
  if c == nil then
    c = const.textcolor;
  end
  local f = drawing.font(fn, fs);
  local t = sprite.text(f, txt, c);
  if copy then
    drawing.copy(t, pic, x, y, alpha);
  else
    sprite.draw(t, pic, x, y);
  end
  sprite.free(t);
end

function drawing.draw_text(pic, txt, x, y, c, fn, fs)
  drawing.render_text(false, pic, txt, x, y, c, fn, fs);
end

function drawing.copy_text(pic, txt, x, y, c, fn, fs, alpha)
  drawing.render_text(true, pic, txt, x, y, c, fn, fs, alpha);
end

function drawing.load_pic(file, cache)
  if not cache then
    return sprite.load(file);
  else
    return drawing.sprite(file);
  end
end

function drawing.draw_pic(pic, file, x, y, cache)
  local p = drawing.load_pic(file, cache);
  sprite.draw(p, pic, x, y);
  if not cache then
    sprite.free(p);
  end
end

function drawing.copy_pic(pic, file, x, y, cache, alpha)
  local p = drawing.load_pic(file, cache);
  drawing.copy(p, pic, x, y, alpha);
  if not cache then
    sprite.free(p);
  end
end

--High level functions
function drawing.textframe(pic, txt, x, y, padx, pady, shiftx, fn, fs, col)
  notnil(x,"x");notnil(y,"y");
  local ss = txt:split("^");
  local arr = {};
  local len = 0;
  for s in ss do
    len = len + 1;
    table.insert(arr, s);
  end
  local tw = 0;
  local linepad = 2;
  for i,v in ipairs(arr) do
    local tw2 = drawing.text_width(v, fn, fs);
    if tw2 > tw then
      tw = tw2;
    end
  end
  local h = drawing.text_height(fn, fs)*len + pady*2 + linepad*(len-1);
  local mw = tw + padx*2;
  local rx = x;
  if shiftx then
    rx = x - mw;
  end  
  local bx = sprite.box(mw, h, const.maskcolor, const.opacity);
  sprite.copy(bx, pic, rx, y);
  sprite.free(bx);  
  
  for i,v in ipairs(arr) do
    local tx = sprite.text(drawing.font(fn, fs), v, col);
    local lh = drawing.text_height(fn, fs);
    sprite.compose(tx, 0, 0, mw, lh, pic, rx+padx, y+pady + (lh+linepad)*(i-1));
    sprite.free(tx);
  end
  
  return mw,h;
end

function drawing.generic_arrow(copy, nav, pic, arr, x, y, sel)
  if nav[arr] ~= nil then
    local file = const.path.ui;
    if not sel then
      file = file..arr..".png";
    else
      file = file..arr.."_sel.png";
    end
    if copy then
      drawing.copy_pic(pic, file, x, y, true);
    else
      drawing.draw_pic(pic, file, x, y, true);
    end
  end
end

function drawing.draw_arrow(nav, pic, arr, x, y, sel)
  drawing.generic_arrow(false, nav, pic, arr, x, y, sel);
end

function drawing.copy_arrow(nav, pic, arr, x, y, sel)
  drawing.generic_arrow(true, nav, pic, arr, x, y, sel);
end

function drawing.draw_portrait(pic, file, x, y, title, titleshift, padx, pady, fn, fs)
  local p = drawing.load_pic(file, false);
  local w,_ = sprite.size(p);
  sprite.draw(p, pic, x, y);
  sprite.free(p);
  if title ~= nil then
    drawing.textframe(pic, title, x + w + titleshift, y, padx, pady, false, fn, fs);
  end
  return pic;
end